#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <openssl/sha.h>
#include <dirent.h>
#include <time.h>

/* TODO:
 * - add subdirectories recursively
*/

void help_message();
int create_announce( const char*, const char*, const char*, const char*, const char*, int );
int create_from_file( const char*, FILE*, unsigned, int );
int create_from_directory( const char*, FILE*, int );
void write_name( const char*, FILE* );

const char* __version__ = "1.0.0";

char* comment = NULL;

int main( int argc, char** argv )
{
	int i;
	int optidx = 0;
	char* announce = NULL;
	char* src = NULL;
	char* outputfile = NULL;
	char* port = "6881";
	char* path = "/announce";
	int piecelen = 256 * 1024;
	while( 1 ) {
		static struct option options[] = {
			{ "announce", 1, 0, 'a' },  // announce
			{ "help", 0, 0, 'h' },
			{ "version", 0, 0, 'v' },
			{ "port", 1, 0, 'p' },
			{ "path", 1, 0, 't' },
			{ "piecelength", 1, 0, 'l' },
			{ "comment", 1, 0, 'c' },
			{ 0, 0, 0, 0 }
		};
		char c = getopt_long( argc, argv, "a:hvp:t:l:c:", options, &optidx );
		if( c == -1 ) {
			break;
		}
		switch( c ) {
			case '?':
				return 1;
			case 'c':
				comment = optarg;
				break;
			case 't':
				path = optarg;
				break;
			case 'a':
				announce = optarg;
				break;
			case 'p':
				port = optarg;
				break;
			case 'h':
				help_message();
				return 0;
			case 'v':
				printf( "createtorrent %s (Daniel Etzold <detzold@gmx.de>)\n", __version__ );
				return 0;
			case 'l':
				piecelen = atoi( optarg );
				break;
		}
	}
	for( i = optind; i < argc; ++i ) {
		if( ! src ) {
			src = argv[ i ];
		} else {
			outputfile = argv[ i ];
		}
	}
	if( announce && src && outputfile && port && path ) {
		return create_announce( announce, src, outputfile, port, path, piecelen );
	}
	fprintf( stderr, "Invalid arguments. Use -h for help.\n" );
	return 1;
}

int create_announce( const char* announce, const char* src, const char* output, const char* port, const char* path, int piecelen )
{
	int ret = 0;
	int p = atoi( port );
	if( p ) {
		FILE* f = fopen( output, "w" );
		if( f ) {
			fprintf( f, "d8:announce%d:%s:%s%s", strlen( announce ) + strlen( path ) + strlen( port ) + 1, announce, port, path );
			fprintf( f, "4:infod12:piece lengthi%de", piecelen );
			if( comment ) {
				fprintf( f, "7:comment%d:%s", strlen( comment ), comment );
			}
			fprintf( f, "13:creation datei%de", (unsigned) time( NULL ) );
			struct stat s;
			if( ! stat( src, &s ) ) {
				if( S_ISDIR( s.st_mode ) ) {
					ret = create_from_directory( src, f, piecelen );
				} else if( S_ISREG( s.st_mode ) ) {
					ret = create_from_file( src, f, s.st_size, piecelen );
				} else {
					fprintf( stderr, "not a file or directory: %s\n", src );
					ret = 1;
				}
			} else {
				fprintf( stderr, "could not access %s: %m\n", src );
				ret = 1;
			}
			fwrite( "ee", 1, 2, f );
			fclose( f );
			if( ret ) {
				unlink( output );
			}
		} else {
			fprintf( stderr, "could not open destination file %s\n", output );
			ret = 1;
		}
	} else {
		fprintf( stderr, "invalid port number\n" );
		ret = 1;
	}
	return ret;
}

void write_name( const char* name, FILE* f )
{
	const char* p = name + strlen( name ) - 1;
	const char* l;
	char* c;
	while( *p == '/' && p != name ) --p;
	l = p;
	while( p != name && *( p - 1 ) != '/' ) --p;
	c = strdup( p );
	c[ l - p + 1 ] = 0;
	fprintf( f, "4:name%d:%s", l - p + 1, c );
	free( c );
}
	
int create_from_directory( const char* src, FILE* f, int piecelen )
{
	int ret = 0;
	DIR* d = opendir( src );
	write_name( src, f );
	if( d ) {
		int files = 0;
		struct dirent* n;
		char* sha = NULL;
		int shasize = 0;
		int bytesin = 0;
		char* buf = (char*) malloc( piecelen );
		fprintf( f, "5:filesl" );
		while( ( n = readdir( d ) ) ) {
			struct stat s;
			char fbuf[ 8092 ];
			sprintf( fbuf, "%s/%s", src, n->d_name );
			if( ! stat( fbuf, &s ) ) {
				if( S_ISREG( s.st_mode ) ) {
					files++;
					FILE* fs = fopen( fbuf, "rb" );
					printf( "adding %s\n", fbuf );
					if( fs ) {
						fprintf( f, "d6:lengthi%de4:pathl%d:%see", (unsigned) s.st_size, strlen( n->d_name ), n->d_name );
						int r;
						while( ( r = fread( buf + bytesin, 1, piecelen - bytesin, fs ) ) ) {
							bytesin += r;
							if( bytesin == piecelen ) {
								sha = (char*) realloc( sha, shasize + 20 );
								SHA1( buf, bytesin, sha + shasize );
								shasize += 20;
								bytesin = 0;
							}
						}
						fclose( fs );
					} else {
						fprintf( stderr, "could not open file %s\n", fbuf );
						ret = 1;
						break;
					}
				} else {
					printf( "ignoring %s\n", fbuf );
				}
			} else {
				fprintf( stderr, "could not stat file %s\n", fbuf );
			}
		}
		fprintf( f, "e" );
		if( shasize || bytesin ) {
			fprintf( f, "6:pieces%d:", shasize + ( ( bytesin > 0 ) ? 20 : 0 ) ); // first e comes from file list
			if( sha ) {
				fwrite( sha, 1, shasize, f );
			}
			if( bytesin ) {
				char sha[ 20 ];
				SHA1( buf, bytesin, sha );
				fwrite( sha, 1, 20, f );
			}
		}
		if( sha ) {
			free( sha );
		}
		free( buf );
		if( ! files ) {
			fprintf( stderr, "warning: no files found in directory\n" );
		}
		closedir( d );
	} else {
		fprintf( stderr, "could not read directory %s\n", src );
		ret = 1;
	}
	return ret;
}

int create_from_file( const char* src, FILE* f, unsigned fsize, int piecelen )
{
	FILE* fs = fopen( src, "rb" );
	write_name( src, f );
	fprintf( f, "6:lengthi%ue", fsize );
	fprintf( f, "6:pieces%d:", ( fsize / piecelen + 1 ) * 20  );
	if( fs ) {
		char sha[ SHA_DIGEST_LENGTH ];
		int r;
		char* buf = (char*) malloc( piecelen );
		printf( "computing sha1... " );
		fflush( stdout );
		while( ( r = fread( buf, 1, piecelen, fs ) ) ) {
			SHA1( buf, r, sha );
			fwrite( sha, 1, 20, f );
		}
		free( buf );
		fclose( fs );
		printf( "done\n" );
	} else {
		fprintf( stderr, "could not open file %s\n", src );
		return 1;
	}
	return 0;
}

void help_message()
{
	printf( "Usage: createtorrent [OPTIONS] -a announce inputfile/directory output\n\n"
		"options:\n"
		"--announce    -a  <announceurl> : announce url\n"
		"--port        -p  <port>        : sets the port, default: 6881\n"
		"--path        -t  <path>        : sets the path on the server, default: /announce\n"
		"--piecelength -l  <piecelen>    : sets the piece length in bytes, default: 262144\n"
		"--comment     -c  <comment>     : adds an optional comment to the torrent file\n"
		"--version     -v                : version of createtorrent\n"
		"--help        -h                : this help screen\n"
	);
}

